import pandas as pd
import re
from datetime import datetime
import os
from uuid import UUID
from typing import Dict, Optional
import pyarrow.dataset as ds
import pyarrow.parquet as pq
from sempy_labs._helper_functions import (
    create_abfss_path,
    save_as_delta_table,
    _get_column_aggregate,
    _create_dataframe,
    _update_dataframe_datatypes,
    resolve_workspace_name_and_id,
    resolve_lakehouse_name_and_id,
    _read_delta_table,
    _mount,
    _create_spark_session,
)
from sempy._utils._log import log
from sempy_labs.lakehouse._get_lakehouse_tables import get_lakehouse_tables
from sempy_labs.lakehouse._lakehouse import lakehouse_attached
import sempy_labs._icons as icons
from tqdm.auto import tqdm


def get_parquet_file_infos(path):

    import notebookutils

    files = []
    items = notebookutils.fs.ls(path)
    for item in items:
        if item.isDir:
            # Ignore the _delta_log directory
            if "_delta_log" not in item.path:
                files.extend(get_parquet_file_infos(item.path))
        else:
            # Filter out non-Parquet files and files with size 0
            if item.path.endswith(".parquet") and item.size > 0:
                files.append((item.path, item.size))
    return files


@log
def delta_analyzer(
    table_name: str,
    approx_distinct_count: bool = True,
    export: bool = False,
    lakehouse: Optional[str | UUID] = None,
    workspace: Optional[str | UUID] = None,
    column_stats: bool = True,
    skip_cardinality: bool = True,
    schema: Optional[str] = None,
) -> Dict[str, pd.DataFrame]:
    """
    Analyzes a delta table and shows the results in dictionary containing a set of 5 dataframes. If 'export' is set to True, the results will be saved to delta tables in the lakehouse attached to the notebook.

    The 5 dataframes returned by this function are:

    * Summary
    * Parquet Files
    * Row Groups
    * Column Chunks
    * Columns

    Read more about Delta Analyzer `here <https://github.com/microsoft/Analysis-Services/tree/master/DeltaAnalyzer>`_.

    Parameters
    ----------
    table_name : str
        The delta table name.
    approx_distinct_count: bool, default=True
        If True, uses approx_count_distinct to calculate the cardinality of each column. If False, uses COUNT(DISTINCT) instead.
    export : bool, default=False
        If True, exports the resulting dataframes to delta tables in the lakehouse attached to the notebook.
    lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse name or ID.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    column_stats : bool, default=True
        If True, collects data about column chunks and columns. If False, skips that step and only returns the other 3 dataframes.
    skip_cardinality : bool, default=True
        If True, skips the cardinality calculation for each column. If False, calculates the cardinality for each column.
    schema : str, default=None
        The name of the schema to which the table belongs (for schema-enabled lakehouses). If None, the default schema is used.

    Returns
    -------
    Dict[str, pandas.DataFrame]
        A dictionary of pandas dataframes showing semantic model objects which violated the best practice analyzer rules.
    """

    # Must calculate column stats if calculating cardinality
    if not skip_cardinality:
        column_stats = True

    prefix = "SLL_DeltaAnalyzer_"
    now = datetime.now()
    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace=workspace)
    (lakehouse_name, lakehouse_id) = resolve_lakehouse_name_and_id(
        lakehouse=lakehouse, workspace=workspace
    )

    delta_table_path = create_abfss_path(
        lakehouse_id, workspace_id, table_name, schema=schema
    )
    local_path = _mount(lakehouse=lakehouse, workspace=workspace)
    if schema is not None:
        table_path = f"{local_path}/Tables/{schema}/{table_name}"
    else:
        table_path = f"{local_path}/Tables/{table_name}"

    parquet_file_df_columns = {
        # "Dataset": "string",
        "Parquet File": "string",
        "Row Count": "int",
        "Row Groups": "int",
        "Created By": "string",
        "Total Table Rows": "int",
        "Total Table Row Groups": "int",
    }
    row_group_df_columns = {
        # "Dataset": "string",
        "Parquet File": "string",
        "Row Group ID": "int",
        "Row Count": "int",
        "Compressed Size": "int",
        "Uncompressed Size": "int",
        "Compression Ratio": "float",
        "Total Table Rows": "int",
        "Ratio Of Total Table Rows": "float",
        "Total Table Row Groups": "int",
    }
    column_chunk_df_columns = {
        # "Dataset": "string",
        "Parquet File": "string",
        "Column ID": "int",
        "Column Name": "string",
        "Column Type": "string",
        "Compressed Size": "int",
        "Uncompressed Size": "int",
        "Has Dict": "bool",
        "Dict Offset": "int_fillna",
        "Value Count": "int",
        "Encodings": "string",
        "Statistics": "string",
        "Primative Type": "string",
    }

    parquet_file_df = _create_dataframe(columns=parquet_file_df_columns)
    row_group_df = _create_dataframe(columns=row_group_df_columns)
    column_chunk_df = _create_dataframe(columns=column_chunk_df_columns)

    row_groups = 0
    max_rows_per_row_group = 0
    min_rows_per_row_group = float("inf")

    schema = ds.dataset(table_path).schema.metadata
    is_vorder = any(b"vorder" in key for key in schema.keys())

    # Get the common details of the Delta table
    spark = _create_spark_session()

    from delta import DeltaTable

    delta_table = DeltaTable.forPath(spark, delta_table_path)
    table_df = delta_table.toDF()
    # total_partition_count = table_df.rdd.getNumPartitions()
    row_count = table_df.count()
    table_details = delta_table.detail().collect()[0].asDict()
    # created_at = table_details.get("createdAt")
    # last_modified = table_details.get("lastModified")
    # partition_columns = table_details.get("partitionColumns")
    # clustering_columns = table_details.get("clusteringColumns")
    num_latest_files = table_details.get("numFiles", 0)
    # size_in_bytes = table_details.get("sizeInBytes")
    # min_reader_version = table_details.get("minReaderVersion")
    # min_writer_version = table_details.get("minWriterVersion")

    latest_files = _read_delta_table(delta_table_path).inputFiles()
    # file_paths = [f.split("/")[-1] for f in latest_files]
    all_parquet_files = get_parquet_file_infos(delta_table_path)
    common_file_paths = set(
        [file_info[0] for file_info in all_parquet_files]
    ).intersection(set(latest_files))
    latest_version_files = [
        file_info
        for file_info in all_parquet_files
        if file_info[0] in common_file_paths
    ]

    for idx, (file_path, file_size) in enumerate(
        bar := tqdm(latest_version_files), start=1
    ):
        file_name = os.path.basename(file_path)
        bar.set_description(
            f"Analyzing the '{file_name}' parquet file ({idx}/{num_latest_files})..."
        )

        relative_path = file_path.split("Tables/")[1]
        file_system_path = f"{local_path}/Tables/{relative_path}"
        parquet_file = pq.ParquetFile(file_system_path)

        row_groups += parquet_file.num_row_groups

        # Generate rowgroup dataframe
        new_data = {
            # "Dataset": "Parquet Files",
            "Parquet File": file_name,
            "Row Count": parquet_file.metadata.num_rows,
            "Row Groups": parquet_file.num_row_groups,
            "Created By": parquet_file.metadata.created_by,
            "Total Table Rows": -1,
            "Total Table Row Groups": -1,
        }

        parquet_file_df = pd.concat(
            [parquet_file_df, pd.DataFrame(new_data, index=[0])], ignore_index=True
        )

        # Loop through the row groups
        for i in range(parquet_file.num_row_groups):
            row_group = parquet_file.metadata.row_group(i)
            num_rows = row_group.num_rows

            max_rows_per_row_group = max(max_rows_per_row_group, num_rows)
            min_rows_per_row_group = min(min_rows_per_row_group, num_rows)

            total_compressed_size = 0
            total_uncompressed_size = 0

            # Loop through the columns
            if column_stats:
                for j in range(row_group.num_columns):
                    column_chunk = row_group.column(j)
                    total_compressed_size += column_chunk.total_compressed_size
                    total_uncompressed_size += column_chunk.total_uncompressed_size

                    # Generate Column Chunk Dataframe
                    new_data = {
                        # "Dataset": "Column Chunks",
                        "Parquet File": file_name,
                        "Column ID": j,
                        "Column Name": column_chunk.path_in_schema,
                        "Column Type": column_chunk.physical_type,
                        "Compressed Size": column_chunk.total_compressed_size,
                        "Uncompressed Size": column_chunk.total_uncompressed_size,
                        "Has Dict": column_chunk.has_dictionary_page,
                        "Dict Offset": column_chunk.dictionary_page_offset,
                        "Value Count": column_chunk.num_values,
                        "Encodings": str(column_chunk.encodings),
                        "Statistics": column_chunk.statistics,
                        "PrimativeType": column_chunk.physical_type,
                    }

                    column_chunk_df = pd.concat(
                        [column_chunk_df, pd.DataFrame(new_data, index=[0])],
                        ignore_index=True,
                    )

            # Generate rowgroup dataframe
            new_data = {
                # "Dataset": "Row Groups",
                "Parquet File": file_name,
                "Row Group ID": i + 1,
                "Row Count": num_rows,
                "Compressed Size": total_compressed_size,
                "Uncompressed Size": total_uncompressed_size,
                "Compression Ratio": (
                    total_compressed_size / total_uncompressed_size
                    if column_stats
                    else 0
                ),
                "Total Table Rows": -1,
                "Total Table Row Groups": -1,
            }

            if not row_group_df.empty:
                row_group_df = pd.concat(
                    [row_group_df, pd.DataFrame(new_data, index=[0])], ignore_index=True
                )
            else:
                row_group_df = pd.DataFrame(new_data, index=[0])

    avg_rows_per_row_group = row_count / row_groups

    # Generate summary dataframe
    summary_df = pd.DataFrame(
        [
            {
                # "Dataset": "Summary",
                "Row Count": row_count,
                "Row Groups": row_groups,
                "Parquet Files": num_latest_files,
                "Max Rows Per Row Group": max_rows_per_row_group,
                "Min Rows Per Row Group": min_rows_per_row_group,
                "Avg Rows Per Row Group": avg_rows_per_row_group,
                "VOrder Enabled": is_vorder,
                # "VOrderLevel": v_order_level,
            }
        ]
    )

    # Clean up data types
    _update_dataframe_datatypes(dataframe=row_group_df, column_map=row_group_df_columns)
    _update_dataframe_datatypes(
        dataframe=parquet_file_df, column_map=parquet_file_df_columns
    )

    # Generate column dataframe
    if column_stats:
        _update_dataframe_datatypes(
            dataframe=column_chunk_df, column_map=column_chunk_df_columns
        )
        column_df = column_chunk_df.groupby(
            ["Column Name", "Column Type"], as_index=False
        ).agg({"Compressed Size": "sum", "Uncompressed Size": "sum"})

        # Add distinct count to column_df
        if not skip_cardinality:
            for ind, r in column_df.iterrows():
                col_name = r["Column Name"]
                if approx_distinct_count:
                    function = "approx"
                else:
                    function = "distinctcount"
                dc = _get_column_aggregate(
                    table_name=table_name,
                    column_name=col_name,
                    function=function,
                    lakehouse=lakehouse,
                    workspace=workspace,
                )

                if "Cardinality" not in column_df.columns:
                    column_df["Cardinality"] = None

                column_df.at[ind, "Cardinality"] = dc

        summary_df["Total Size"] = column_df["Compressed Size"].sum()

    parquet_file_df["Total Table Rows"] = parquet_file_df["Row Count"].sum()
    parquet_file_df["Total Table Row Groups"] = parquet_file_df["Row Groups"].sum()

    row_group_df["Total Table Rows"] = parquet_file_df["Row Count"].sum()
    row_group_df["Total Table Row Groups"] = parquet_file_df["Row Groups"].sum()
    total_rows = row_group_df["Row Count"].sum()
    row_group_df["Ratio Of Total Table Rows"] = (
        row_group_df["Row Count"] / total_rows * 100.0
    )

    if column_stats:
        column_df["Total Table Rows"] = parquet_file_df["Row Count"].sum()
        column_df["Table Size"] = column_df["Compressed Size"].sum()
        column_df["Size Percent Of Table"] = (
            column_df["Compressed Size"] / column_df["Table Size"] * 100.0
        )
    if not skip_cardinality and column_stats:
        column_df["Cardinality"] = column_df["Cardinality"].fillna(0).astype(int)
        column_df["Cardinality Of Total Rows"] = (
            column_df["Cardinality"] / column_df["Total Table Rows"] * 100.0
        )

    dataframes = {
        "Summary": summary_df,
        "Parquet Files": parquet_file_df,
        "Row Groups": row_group_df,
    }

    if column_stats:
        dataframes["Column Chunks"] = column_chunk_df
        dataframes["Columns"] = column_df

    save_table = f"{prefix}Summary"

    if export:
        if not lakehouse_attached():
            raise ValueError(
                f"{icons.red_dot} No lakehouse is attached to this notebook. Please attach a lakehouse to the notebook before running the Delta Analyzer."
            )
        dfL = get_lakehouse_tables()
        dfL_filt = dfL[dfL["Table Name"] == save_table]
        if dfL_filt.empty:
            runId = 1
        else:
            max_run_id = _get_column_aggregate(
                table_name=save_table,
            )
            runId = max_run_id + 1

    for name, df in dataframes.items():
        name = name.replace(" ", "")
        cols = {
            "Workspace Name": workspace_name,
            "Workspace Id": workspace_id,
            "Lakehouse Name": lakehouse_name,
            "Lakehouse Id": lakehouse_id,
            "Table Name": table_name,
        }
        for i, (col, param) in enumerate(cols.items()):
            df[col] = param
            df.insert(i, col, df.pop(col))

        df["Timestamp"] = now
        df["Timestamp"] = pd.to_datetime(df["Timestamp"])

        if export:
            df["Run Id"] = runId
            df["Run Id"] = df["Run Id"].astype(int)

            df.columns = df.columns.str.replace(" ", "")
            save_as_delta_table(
                dataframe=df,
                delta_table_name=f"{prefix}{name}",
                write_mode="append",
                merge_schema=True,
            )

    return dataframes


@log
def get_delta_table_history(
    table_name: str,
    lakehouse: Optional[str | UUID] = None,
    workspace: Optional[str | UUID] = None,
    schema: Optional[str] = None,
) -> pd.DataFrame:
    """
    Returns the history of a delta table as a pandas dataframe.

    Parameters
    ----------
    table_name : str
        The delta table name.
    lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse name or ID.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    schema : str, default=None
        The name of the schema to which the table belongs (for schema-enabled lakehouses). If None, the default schema is used.

    Returns
    -------
    pandas.DataFrame
        A dataframe showing the history of the delta table.
    """

    def camel_to_title(text):
        return re.sub(r"([a-z])([A-Z])", r"\1 \2", text).title()

    spark = _create_spark_session()

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace=workspace)
    (lakehouse_name, lakehouse_id) = resolve_lakehouse_name_and_id(
        lakehouse=lakehouse, workspace=workspace
    )
    path = create_abfss_path(lakehouse_id, workspace_id, table_name, schema)

    from delta import DeltaTable

    delta_table = DeltaTable.forPath(spark, path)
    df = delta_table.history().toPandas()

    df.rename(columns=lambda col: camel_to_title(col), inplace=True)

    return df



import pandas as pd
from typing import Optional
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    _base_api,
    _create_dataframe,
    delete_item,
    create_item,
)
from uuid import UUID
import sempy_labs._icons as icons


def list_kql_databases(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    Shows the KQL databases within a workspace.

    This is a wrapper function for the following API: `Items - List KQL Databases <https://learn.microsoft.com/rest/api/fabric/kqldatabase/items/list-kql-databases>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the KQL databases within a workspace.
    """

    columns = {
        "KQL Database Name": "string",
        "KQL Database Id": "string",
        "Description": "string",
        "Parent Eventhouse Item Id": "string",
        "Query Service URI": "string",
        "Ingestion Service URI": "string",
        "Database Type": "string",
    }
    df = _create_dataframe(columns=columns)

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    responses = _base_api(
        request=f"v1/workspaces/{workspace_id}/kqlDatabases",
        uses_pagination=True,
        client="fabric_sp",
    )

    for r in responses:
        for v in r.get("value", []):
            prop = v.get("properties", {})

            new_data = {
                "KQL Database Name": v.get("displayName"),
                "KQL Database Id": v.get("id"),
                "Description": v.get("description"),
                "Parent Eventhouse Item Id": prop.get("parentEventhouseItemId"),
                "Query Service URI": prop.get("queryServiceUri"),
                "Ingestion Service URI": prop.get("ingestionServiceUri"),
                "Database Type": prop.get("databaseType"),
            }
            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


def _create_kql_database(
    name: str, description: Optional[str] = None, workspace: Optional[str | UUID] = None
):
    """
    Creates a KQL database.

    This is a wrapper function for the following API: `Items - Create KQL Database <https://learn.microsoft.com/rest/api/fabric/kqldatabase/items/create-kql-database>`_.

    Parameters
    ----------
    name: str
        Name of the KQL database.
    description : str, default=None
        A description of the environment.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    create_item(
        name=name, description=description, type="KQLDatabase", workspace=workspace
    )


def delete_kql_database(
    kql_database: str | UUID,
    workspace: Optional[str | UUID] = None,
    **kwargs,
):
    """
    Deletes a KQL database.

    This is a wrapper function for the following API: `Items - Delete KQL Database <https://learn.microsoft.com/rest/api/fabric/kqldatabase/items/delete-kql-database>`_.

    Parameters
    ----------
    kql_database: str | uuid.UUID
        Name or ID of the KQL database.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    if "name" in kwargs:
        kql_database = kwargs["name"]
        print(
            f"{icons.warning} The 'name' parameter is deprecated. Please use 'kql_database' instead."
        )

    delete_item(item=kql_database, type="KQLDatabase", workspace=workspace)



import sempy.fabric as fabric
import pandas as pd
import datetime
from sempy_labs._helper_functions import (
    save_as_delta_table,
    resolve_workspace_capacity,
    retry,
    _get_column_aggregate,
    resolve_workspace_id,
    resolve_lakehouse_name_and_id,
)
from sempy_labs.lakehouse import (
    get_lakehouse_tables,
    lakehouse_attached,
)
from sempy_labs._model_bpa import run_model_bpa
from typing import Optional, List
from sempy._utils._log import log
import sempy_labs._icons as icons
from uuid import UUID


@log
def run_model_bpa_bulk(
    rules: Optional[pd.DataFrame] = None,
    extended: bool = False,
    language: Optional[str] = None,
    workspace: Optional[str | UUID | List[str | UUID]] = None,
    skip_models: Optional[str | List[str]] = ["ModelBPA", "Fabric Capacity Metrics"],
    skip_models_in_workspace: Optional[dict] = None,
):
    """
    Runs the semantic model Best Practice Analyzer across all semantic models in a workspace (or all accessible workspaces).
    Saves (appends) the results to the 'modelbparesults' delta table in the lakehouse attached to the notebook.
    Default semantic models are skipped in this analysis.

    Parameters
    ----------
    rules : pandas.DataFrame, default=None
        A pandas dataframe containing rules to be evaluated. Based on the format of the dataframe produced by the model_bpa_rules function.
    extended : bool, default=False
        If True, runs the set_vertipaq_annotations function to collect Vertipaq Analyzer statistics to be used in the analysis of the semantic model.
    language : str, default=None
        The language (code) in which the rules will appear. For example, specifying 'it-IT' will show the Rule Name, Category and Description in Italian.
        Defaults to None which resolves to English.
    workspace : str | uuid.UUID | List[str | uuid.UUID], default=None
        The workspace or list of workspaces to scan. Supports both the workspace name and the workspace id.
        Defaults to None which scans all accessible workspaces.
    skip_models : str | List[str], default=['ModelBPA', 'Fabric Capacity Metrics']
        The semantic models to always skip when running this analysis.
    skip_models_in_workspace : dict, default=None
        A dictionary showing specific semantic models within specific workspaces to skip. See the example below:
        {
            "Workspace A": ["Dataset1", "Dataset2"],
            "Workspace B": ["Dataset5", "Dataset 8"],
        }
    """

    if not lakehouse_attached():
        raise ValueError(
            f"{icons.red_dot} No lakehouse is attached to this notebook. Must attach a lakehouse to the notebook."
        )

    if isinstance(skip_models, str):
        skip_models = [skip_models]

    skip_models.extend(["ModelBPA", "Fabric Capacity Metrics"])

    now = datetime.datetime.now()
    output_table = "modelbparesults"
    lakeT = get_lakehouse_tables()
    lakeT_filt = lakeT[lakeT["Table Name"] == output_table]
    if lakeT_filt.empty:
        runId = 1
    else:
        max_run_id = _get_column_aggregate(table_name=output_table)
        runId = max_run_id + 1

    if isinstance(workspace, str):
        workspace = [workspace]

    dfW = fabric.list_workspaces("type ne 'AdminInsights'")
    if workspace is None:
        dfW_filt = dfW.copy()
    else:
        dfW_filt = dfW[(dfW["Name"].isin(workspace)) | (dfW["Id"].isin(workspace))]

    if dfW_filt.empty:
        raise ValueError(
            f"{icons.red_dot} There are no valid workspaces to assess. This is likely due to not having proper permissions to the workspace(s) entered in the 'workspace' parameter."
        )

    for _, r in dfW_filt.iterrows():
        wksp = r["Name"]
        wksp_id = r["Id"]
        capacity_id, capacity_name = resolve_workspace_capacity(workspace=wksp)
        df = pd.DataFrame(columns=list(icons.bpa_schema.keys()))
        dfD = fabric.list_datasets(workspace=wksp, mode="rest")

        # Skip models in workspace
        if skip_models_in_workspace is not None and isinstance(
            skip_models_in_workspace, dict
        ):
            skip_models_wkspc = skip_models_in_workspace.get(wksp)
            dfD = dfD[~dfD["Dataset Name"].isin(skip_models_wkspc)]

        # Exclude default semantic models
        if not dfD.empty:
            dfI = fabric.list_items(workspace=wksp)
            filtered_df = dfI.groupby("Display Name").filter(
                lambda x: set(["Warehouse", "SemanticModel"]).issubset(set(x["Type"]))
                or set(["Lakehouse", "SemanticModel"]).issubset(set(x["Type"]))
            )
            default_semantic_models = filtered_df["Display Name"].unique().tolist()
            skip_models.extend(default_semantic_models)
            dfD_filt = dfD[~dfD["Dataset Name"].isin(skip_models)]

            if not dfD_filt.empty:
                for _, r2 in dfD_filt.iterrows():
                    dataset_id = r2["Dataset Id"]
                    dataset_name = r2["Dataset Name"]
                    config_by = r2["Configured By"]
                    print(
                        f"{icons.in_progress} Collecting Model BPA stats for the '{dataset_name}' semantic model within the '{wksp}' workspace."
                    )
                    try:
                        bpa_df = run_model_bpa(
                            dataset=dataset_id,
                            workspace=wksp,
                            language=language,
                            return_dataframe=True,
                            rules=rules,
                            extended=extended,
                        )
                        bpa_df["Capacity Name"] = capacity_name
                        bpa_df["Capacity Id"] = capacity_id
                        bpa_df["Workspace Name"] = wksp
                        bpa_df["Workspace Id"] = wksp_id
                        bpa_df["Dataset Name"] = dataset_name
                        bpa_df["Dataset Id"] = dataset_id
                        bpa_df["Configured By"] = config_by
                        bpa_df["Timestamp"] = now
                        bpa_df["RunId"] = runId
                        bpa_df = bpa_df[list(icons.bpa_schema.keys())]

                        bpa_df["RunId"] = bpa_df["RunId"].astype("int")

                        if df.empty:
                            df = bpa_df
                        elif not bpa_df.empty:
                            df = pd.concat([df, bpa_df], ignore_index=True)
                        print(
                            f"{icons.green_dot} Collected Model BPA stats for the '{dataset_name}' semantic model within the '{wksp}' workspace."
                        )
                    except Exception as e:
                        print(
                            f"{icons.red_dot} Model BPA failed for the '{dataset_name}' semantic model within the '{wksp}' workspace."
                        )
                        print(e)

                if df.empty:
                    print(
                        f"{icons.yellow_dot} No BPA results to save for the '{wksp}' workspace."
                    )
                else:
                    df["Severity"].replace(icons.severity_mapping, inplace=True)

                    # Append save results individually for each workspace (so as not to create a giant dataframe)
                    print(
                        f"{icons.in_progress} Saving the Model BPA results of the '{wksp}' workspace to the '{output_table}' within the lakehouse attached to this notebook..."
                    )

                    schema = {
                        key.replace(" ", "_"): value
                        for key, value in icons.bpa_schema.items()
                    }

                    save_as_delta_table(
                        dataframe=df,
                        delta_table_name=output_table,
                        write_mode="append",
                        schema=schema,
                        merge_schema=True,
                    )
                    print(
                        f"{icons.green_dot} Saved BPA results to the '{output_table}' delta table."
                    )

    print(f"{icons.green_dot} Bulk BPA scan complete.")


@log
def create_model_bpa_semantic_model(
    dataset: Optional[str] = icons.model_bpa_name,
    lakehouse: Optional[str | UUID] = None,
    lakehouse_workspace: Optional[str | UUID] = None,
):
    """
    Dynamically generates a Direct Lake semantic model based on the 'modelbparesults' delta table which contains the Best Practice Analyzer results.
    This semantic model used in combination with the corresponding Best Practice Analyzer report can be used to analyze multiple semantic models
    on multiple workspaces at once (and over time).

    The semantic model is always created within the same workspace as the lakehouse.

    Parameters
    ----------
    dataset : str, default='ModelBPA'
        Name of the semantic model to be created.
    lakehouse : str | uuid.UUID, default=None
        Name of the Fabric lakehouse which contains the 'modelbparesults' delta table.
        Defaults to None which resolves to the default lakehouse attached to the notebook.
    lakehouse_workspace : str | uuid.UUID, default=None
        The workspace in which the lakehouse resides.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    from sempy_labs.directlake import (
        generate_shared_expression,
        add_table_to_direct_lake_semantic_model,
    )
    from sempy_labs import create_blank_semantic_model, refresh_semantic_model
    from sempy_labs.tom import connect_semantic_model

    lakehouse_workspace_id = resolve_workspace_id(workspace=lakehouse_workspace)
    (lakehouse_id, lakehouse_name) = resolve_lakehouse_name_and_id(
        lakehouse=lakehouse, workspace=lakehouse_workspace_id
    )

    # Generate the shared expression based on the lakehouse and lakehouse workspace
    expr = generate_shared_expression(
        item_name=lakehouse_name,
        item_type="Lakehouse",
        workspace=lakehouse_workspace_id,
    )

    # Create blank model
    create_blank_semantic_model(
        dataset=dataset, workspace=lakehouse_workspace_id, overwrite=True
    )

    @retry(
        sleep_time=1,
        timeout_error_message=f"{icons.red_dot} Function timed out after 1 minute",
    )
    def dyn_connect():
        with connect_semantic_model(
            dataset=dataset, readonly=True, workspace=lakehouse_workspace_id
        ) as tom:

            tom.model

    dyn_connect()
    icons.sll_tags.append("ModelBPABulk")
    table_exists = False
    with connect_semantic_model(
        dataset=dataset, readonly=False, workspace=lakehouse_workspace_id
    ) as tom:
        t_name = "BPAResults"
        t_name_full = f"'{t_name}'"
        # Create the shared expression
        if not any(e.Name == "DatabaseQuery" for e in tom.model.Expressions):
            tom.add_expression(name="DatabaseQuery", expression=expr)
        # Add the table to the model
        if any(t.Name == t_name for t in tom.model.Tables):
            table_exists = True
    if not table_exists:
        add_table_to_direct_lake_semantic_model(
            dataset=dataset,
            table_name=t_name,
            lakehouse_table_name="modelbparesults",
            workspace=lakehouse_workspace_id,
            refresh=False,
        )
    with connect_semantic_model(
        dataset=dataset, readonly=False, workspace=lakehouse_workspace_id
    ) as tom:
        # Fix column names
        for c in tom.all_columns():
            if c.Name == "Dataset_Name":
                c.Name = "Model"
            elif c.Name == "Dataset_Id":
                c.Name = "Model Id"
            elif c.Name == "Workspace_Name":
                c.Name = "Workspace"
            elif c.Name == "Capacity_Name":
                c.Name = "Capacity"
            elif c.Name == "Configured_By":
                c.Name = "Model Owner"
            elif c.Name == "URL":
                c.DataCategory = "WebURL"
            elif c.Name == "RunId":
                tom.set_summarize_by(
                    table_name=c.Parent.Name, column_name=c.Name, value="None"
                )
            c.Name = c.Name.replace("_", " ")

        # Implement pattern for base measures
        def get_expr(table_name, calculation):
            return f"IF(HASONEFILTER({table_name}[RunId]),{calculation},CALCULATE({calculation},FILTER(VALUES({table_name}[RunId]),{table_name}[RunId] = [Max Run Id])))"

        # Add measures
        int_format = "#,0"
        m_name = "Max Run Id"
        if not any(m.Name == m_name for m in tom.all_measures()):
            tom.add_measure(
                table_name=t_name,
                measure_name=m_name,
                expression=f"CALCULATE(MAX({t_name_full}[RunId]),{t_name_full}[RunId])",
                format_string=int_format,
            )
        m_name = "Capacities"
        if not any(m.Name == m_name for m in tom.all_measures()):
            calc = f"COUNTROWS(DISTINCT({t_name_full}[Capacity]))"
            tom.add_measure(
                table_name=t_name,
                measure_name=m_name,
                expression=get_expr(t_name_full, calc),
                format_string=int_format,
            )
        m_name = "Models"
        if not any(m.Name == m_name for m in tom.all_measures()):
            calc = f"COUNTROWS(DISTINCT({t_name_full}[Model]))"
            tom.add_measure(
                table_name=t_name,
                measure_name=m_name,
                expression=get_expr(t_name_full, calc),
                format_string=int_format,
            )
        m_name = "Workspaces"
        if not any(m.Name == m_name for m in tom.all_measures()):
            calc = f"COUNTROWS(DISTINCT({t_name_full}[Workspace]))"
            tom.add_measure(
                table_name=t_name,
                measure_name=m_name,
                expression=get_expr(t_name_full, calc),
                format_string=int_format,
            )
        m_name = "Violations"
        if not any(m.Name == m_name for m in tom.all_measures()):
            calc = f"COUNTROWS({t_name_full})"
            tom.add_measure(
                table_name=t_name,
                measure_name=m_name,
                expression=get_expr(t_name_full, calc),
                format_string=int_format,
            )
        m_name = "Error Violations"
        if not any(m.Name == m_name for m in tom.all_measures()):
            tom.add_measure(
                table_name=t_name,
                measure_name=m_name,
                expression=f'CALCULATE([Violations],{t_name_full}[Severity]="Error")',
                format_string=int_format,
            )
        m_name = "Rules Violated"
        if not any(m.Name == m_name for m in tom.all_measures()):
            calc = f"COUNTROWS(DISTINCT({t_name_full}[Rule Name]))"
            tom.add_measure(
                table_name=t_name,
                measure_name=m_name,
                expression=get_expr(t_name_full, calc),
                format_string=int_format,
            )
        m_name = "Rule Severity"
        if not any(m.Name == m_name for m in tom.all_measures()):
            tom.add_measure(
                table_name=t_name,
                measure_name=m_name,
                expression=f"IF(ISFILTERED({t_name_full}[Rule Name]),IF( HASONEVALUE({t_name_full}[Rule Name]),MIN({t_name_full}[Severity])))",
            )
        # tom.add_measure(table_name=t_name, measure_name='Rules Followed', expression="[Rules] - [Rules Violated]")

    # Refresh the model
    refresh_semantic_model(dataset=dataset, workspace=lakehouse_workspace_id)



import pandas as pd
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    _is_valid_uuid,
    _update_dataframe_datatypes,
    _base_api,
    _create_dataframe,
    resolve_workspace_name,
)
from typing import Optional, Tuple
import sempy_labs._icons as icons
from uuid import UUID


def list_dataflows(workspace: Optional[str | UUID] = None):
    """
    Shows a list of all dataflows which exist within a workspace.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the dataflows which exist within a workspace.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    columns = {
        "Dataflow Id": "string",
        "Dataflow Name": "string",
        "Configured By": "string",
        "Users": "string",
        "Generation": "int",
    }
    df = _create_dataframe(columns=columns)

    response = _base_api(request=f"/v1.0/myorg/groups/{workspace_id}/dataflows")

    data = []  # Collect rows here

    for v in response.json().get("value", []):
        new_data = {
            "Dataflow Id": v.get("objectId"),
            "Dataflow Name": v.get("name"),
            "Configured By": v.get("configuredBy"),
            "Users": v.get("users", []),
            "Generation": v.get("generation"),
        }
        data.append(new_data)

    if data:
        df = pd.DataFrame(data)

        _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


def assign_workspace_to_dataflow_storage(
    dataflow_storage_account: str, workspace: Optional[str | UUID] = None
):
    """
    Assigns a dataflow storage account to a workspace.

    This is a wrapper function for the following API: `Dataflow Storage Accounts - Groups AssignToDataflowStorage <https://learn.microsoft.com/rest/api/power-bi/dataflow-storage-accounts/groups-assign-to-dataflow-storage>`_.

    Parameters
    ----------
    dataflow_storage_account : str
        The name of the dataflow storage account.
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    df = list_dataflow_storage_accounts()
    df_filt = df[df["Dataflow Storage Account Name"] == dataflow_storage_account]

    if len(df_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{dataflow_storage_account}' does not exist."
        )

    dataflow_storage_id = df_filt["Dataflow Storage Account ID"].iloc[0]
    payload = {"dataflowStorageId": dataflow_storage_id}

    _base_api(
        request=f"/v1.0/myorg/groups/{workspace_id}/AssignToDataflowStorage",
        method="post",
        payload=payload,
    )

    print(
        f"{icons.green_dot} The '{dataflow_storage_account}' dataflow storage account has been assigned to the '{workspace_name}' workspacce."
    )


def list_dataflow_storage_accounts() -> pd.DataFrame:
    """
    Shows the accessible dataflow storage accounts.

    This is a wrapper function for the following API: `Dataflow Storage Accounts - Get Dataflow Storage Accounts <https://learn.microsoft.com/rest/api/power-bi/dataflow-storage-accounts/get-dataflow-storage-accounts>`_.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the accessible dataflow storage accounts.
    """

    columns = {
        "Dataflow Storage Account ID": "string",
        "Dataflow Storage Account Name": "string",
        "Enabled": "bool",
    }
    df = _create_dataframe(columns=columns)

    response = _base_api(request="/v1.0/myorg/dataflowStorageAccounts")

    for v in response.json().get("value", []):
        new_data = {
            "Dataflow Storage Account ID": v.get("id"),
            "Dataflow Storage Account Name": v.get("name"),
            "Enabled": v.get("isEnabled"),
        }
        df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


def list_upstream_dataflows(
    dataflow: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Shows a list of upstream dataflows for the specified dataflow.

    This is a wrapper function for the following API: `Dataflows - Get Upstream Dataflows In Group <https://learn.microsoft.com/rest/api/power-bi/dataflows/get-upstream-dataflows-in-group>`_.

    Parameters
    ----------
    dataflow : str | uuid.UUID
        Name or UUID of the dataflow.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of upstream dataflows for the specified dataflow.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataflow_name, dataflow_id) = _resolve_dataflow_name_and_id(
        dataflow=dataflow, workspace=workspace_id
    )

    columns = {
        "Dataflow Name": "string",
        "Dataflow Id": "string",
        "Workspace Name": "string",
        "Workspace Id": "string",
        "Upstream Dataflow Name": "string",
        "Upstream Dataflow Id": "string",
        "Upstream Workspace Name": "string",
        "Upstream Workspace Id": "string",
    }
    df = _create_dataframe(columns=columns)

    def collect_upstreams(dataflow_id, dataflow_name, workspace_id, workspace_name):
        response = _base_api(
            request=f"/v1.0/myorg/groups/{workspace_id}/dataflows/{dataflow_id}/upstreamDataflows"
        )

        values = response.json().get("value", [])
        for v in values:
            tgt_dataflow_id = v.get("targetDataflowId")
            tgt_workspace_id = v.get("groupId")
            tgt_workspace_name = resolve_workspace_name(workspace_id=tgt_workspace_id)
            (tgt_dataflow_name, _) = _resolve_dataflow_name_and_id(
                dataflow=tgt_dataflow_id, workspace=tgt_workspace_id
            )

            df.loc[len(df)] = {
                "Dataflow Name": dataflow_name,
                "Dataflow Id": dataflow_id,
                "Workspace Name": workspace_name,
                "Workspace Id": workspace_id,
                "Upstream Dataflow Name": tgt_dataflow_name,
                "Upstream Dataflow Id": tgt_dataflow_id,
                "Upstream Workspace Name": tgt_workspace_name,
                "Upstream Workspace Id": tgt_workspace_id,
            }

            collect_upstreams(
                tgt_dataflow_id,
                tgt_dataflow_name,
                tgt_workspace_id,
                tgt_workspace_name,
            )

    collect_upstreams(dataflow_id, dataflow_name, workspace_id, workspace_name)

    return df


def _resolve_dataflow_name_and_id(
    dataflow: str | UUID, workspace: Optional[str | UUID] = None
) -> Tuple[str, UUID]:

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    dfD = list_dataflows(workspace=workspace_id)

    if _is_valid_uuid(dataflow):
        dfD_filt = dfD[dfD["Dataflow Id"] == dataflow]
    else:
        dfD_filt = dfD[dfD["Dataflow Name"] == dataflow]

    if len(dfD_filt) == 0:
        raise ValueError(
            f"{icons.red_dot} The '{dataflow}' dataflow does not exist within the '{workspace_name}' workspace."
        )

    dataflow_id = dfD_filt["Dataflow Id"].iloc[0]
    dataflow_name = dfD_filt["Dataflow Name"].iloc[0]

    return dataflow_name, dataflow_id



from sempy_labs._helper_functions import (
    resolve_workspace_id,
    resolve_lakehouse_id,
    _xml_to_dict,
    _create_dataframe,
    _update_dataframe_datatypes,
)
from sempy._utils._log import log
from uuid import UUID
from typing import Optional, List
import sempy_labs._icons as icons
import xml.etree.ElementTree as ET
import pandas as pd


def _request_blob_api(
    request: str,
    method: str = "get",
    payload: Optional[dict] = None,
    status_codes: int | List[int] = 200,
):

    import requests
    import notebookutils
    from sempy.fabric.exceptions import FabricHTTPException

    if isinstance(status_codes, int):
        status_codes = [status_codes]

    token = notebookutils.credentials.getToken("storage")

    headers = {
        "Authorization": f"Bearer {token}",
        "Content-Type": "application/json",
        "x-ms-version": "2025-05-05",
    }

    response = requests.request(
        method.upper(),
        f"https://onelake.blob.fabric.microsoft.com/{request}",
        headers=headers,
        json=payload,
    )

    if response.status_code not in status_codes:
        raise FabricHTTPException(response)

    return response


@log
def list_blobs(
    lakehouse: Optional[str | UUID] = None,
    workspace: Optional[str | UUID] = None,
    container: Optional[str] = None,
) -> pd.DataFrame:
    """
    Returns a list of blobs for a given lakehouse.

    This function leverages the following API: `List Blobs <https://learn.microsoft.com/rest/api/storageservices/list-blobs?tabs=microsoft-entra-id>`_.

    Parameters
    ----------
    lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse name or ID.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    container : str, default=None
        The container name to list blobs from. If None, lists all blobs in the lakehouse.
        Valid values are "Tables" or "Files". If not specified, the function will list all blobs in the lakehouse.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of blobs in the lakehouse.
    """

    workspace_id = resolve_workspace_id(workspace)
    lakehouse_id = resolve_lakehouse_id(lakehouse, workspace_id)

    if container is None:
        path_prefix = f"{workspace_id}/{lakehouse_id}"
    else:
        if container not in ["Tables", "Files"]:
            raise ValueError(
                f"{icons.red_dot} Invalid container '{container}' within the file_path parameter. Expected 'Tables' or 'Files'."
            )
        path_prefix = f"{workspace_id}/{lakehouse_id}/{container}"

    response = _request_blob_api(
        request=f"{path_prefix}?restype=container&comp=list&include=deleted"
    )
    root = ET.fromstring(response.content)
    response_json = _xml_to_dict(root)

    columns = {
        "Blob Name": "str",
        "Is Deleted": "bool",
        "Deletion Id": "str",
        "Creation Time": "datetime",
        "Expiry Time": "datetime",
        "Etag": "str",
        "Resource Type": "str",
        "Content Length": "int",
        "Content Type": "str",
        "Content Encoding": "str",
        "Content Language": "str",
        "Content CRC64": "str",
        "Content MD5": "str",
        "Cache Control": "str",
        "Content Disposition": "str",
        "Blob Type": "str",
        "Access Tier": "str",
        "Access Tier Inferred": "str",
        "Server Encrypted": "bool",
        "Deleted Time": "str",
        "Remaining Retention Days": "str",
    }

    df = _create_dataframe(columns=columns)

    for blob in (
        response_json.get("EnumerationResults", {}).get("Blobs", {}).get("Blob", {})
    ):
        p = blob.get("Properties", {})
        new_data = {
            "Blob Name": blob.get("Name"),
            "Is Deleted": blob.get("Deleted", False),
            "Deletion Id": blob.get("DeletionId"),
            "Creation Time": p.get("Creation-Time"),
            "Expiry Time": p.get("Expiry-Time"),
            "Etag": p.get("Etag"),
            "Resource Type": p.get("ResourceType"),
            "Content Length": p.get("Content-Length"),
            "Content Type": p.get("Content-Type"),
            "Content Encoding": p.get("Content-Encoding"),
            "Content Language": p.get("Content-Language"),
            "Content CRC64": p.get("Content-CRC64"),
            "Content MD5": p.get("Content-MD5"),
            "Cache Control": p.get("Cache-Control"),
            "Content Disposition": p.get("Content-Disposition"),
            "Blob Type": p.get("BlobType"),
            "Access Tier": p.get("AccessTier"),
            "Access Tier Inferred": p.get("AccessTierInferred"),
            "Server Encrypted": p.get("ServerEncrypted"),
            "Deleted Time": p.get("DeletedTime"),
            "Remaining Retention Days": p.get("RemainingRetentionDays"),
        }

        df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


@log
def recover_lakehouse_object(
    file_path: str,
    lakehouse: Optional[str | UUID] = None,
    workspace: Optional[str | UUID] = None,
):
    """
    Recovers an object (i.e. table, file, folder) in a lakehouse from a deleted state. Only `soft-deleted objects <https://learn.microsoft.com/fabric/onelake/onelake-disaster-recovery#soft-delete-for-onelake-files>`_ can be recovered (deleted for less than 7 days).

    Parameters
    ----------
    file_path : str
        The file path of the object to restore. For example: "Tables/my_delta_table".
    lakehouse : str | uuid.UUID, default=None
        The Fabric lakehouse name or ID.
        Defaults to None which resolves to the lakehouse attached to the notebook.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID used by the lakehouse.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    workspace_id = resolve_workspace_id(workspace)
    lakehouse_id = resolve_lakehouse_id(lakehouse, workspace_id)

    blob_path_prefix = f"{lakehouse_id}/{file_path}"

    container = file_path.split("/")[0]
    if container not in ["Tables", "Files"]:
        raise ValueError(
            f"{icons.red_dot} Invalid container '{container}' within the file_path parameter. Expected 'Tables' or 'Files'."
        )

    df = list_blobs(lakehouse=lakehouse, workspace=workspace, container=container)

    for _, r in df.iterrows():
        blob_name = r.get("Blob Name")
        is_deleted = r.get("Is Deleted")
        if blob_name.startswith(blob_path_prefix) and is_deleted:
            print(f"{icons.in_progress} Restoring the '{blob_name}' blob...")
            _request_blob_api(
                request=f"{workspace_id}/{lakehouse_id}/{file_path}?comp=undelete",
                method="put",
            )
            print(f"{icons.green_dot} The '{blob_name}' blob has been restored.")


def _get_user_delegation_key():

    # https://learn.microsoft.com/rest/api/storageservices/get-user-delegation-key

    from datetime import datetime, timedelta, timezone

    utc_now = datetime.now(timezone.utc)
    start_time = utc_now + timedelta(minutes=2)
    expiry_time = start_time + timedelta(minutes=45)
    start_str = start_time.strftime("%Y-%m-%dT%H:%M:%SZ")
    expiry_str = expiry_time.strftime("%Y-%m-%dT%H:%M:%SZ")

    payload = f"""<?xml version="1.0" encoding="utf-8"?>
    <KeyInfo>
        <Start>{start_str}</Start>
        <Expiry>{expiry_str}</Expiry>
    </KeyInfo>"""

    response = _request_blob_api(
        request="restype=service&comp=userdelegationkey",
        method="post",
        payload=payload,
    )

    return response.content



from uuid import UUID
from typing import Optional
import pandas as pd
from sempy_labs._helper_functions import (
    _create_dataframe,
    _base_api,
    _update_dataframe_datatypes,
    resolve_workspace_name_and_id,
    resolve_dataset_name_and_id,
    delete_item,
)
import sempy_labs._icons as icons


def get_semantic_model_refresh_schedule(
    dataset: str | UUID, workspace: Optional[str | UUID] = None
) -> pd.DataFrame:
    """
    Gets the refresh schedule for the specified dataset from the specified workspace.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        Shows the refresh schedule for the specified dataset from the specified workspace.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace)

    columns = {
        "Days": "str",
        "Times": "str",
        "Enabled": "bool",
        "Local Time Zone Id": "str",
        "Notify Option": "str",
    }

    column_map = {
        "days": "Days",
        "times": "Times",
        "enabled": "Enabled",
        "localTimeZoneId": "Local Time Zone Id",
        "notifyOption": "Notify Option",
    }

    df = _create_dataframe(columns)

    result = _base_api(
        request=f"/v1.0/myorg/groups/{workspace_id}/datasets/{dataset_id}/refreshSchedule"
    ).json()

    df = (
        pd.json_normalize(result)
        .drop(columns=["@odata.context"], errors="ignore")
        .rename(columns=column_map)
    )

    _update_dataframe_datatypes(dataframe=df, column_map=columns)

    return df


def enable_semantic_model_scheduled_refresh(
    dataset: str | UUID,
    workspace: Optional[str | UUID] = None,
    enable: bool = True,
):
    """
    Enables the scheduled refresh for the specified dataset from the specified workspace.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    enable : bool, default=True
        If True, enables the scheduled refresh.
        If False, disables the scheduled refresh.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace)

    df = get_semantic_model_refresh_schedule(dataset=dataset, workspace=workspace)
    status = df["Enabled"].iloc[0]

    if enable and status:
        print(
            f"{icons.info} Scheduled refresh for the '{dataset_name}' within the '{workspace_name}' workspace is already enabled."
        )
    elif not enable and not status:
        print(
            f"{icons.info} Scheduled refresh for the '{dataset_name}' within the '{workspace_name}' workspace is already disabled."
        )
    else:
        payload = {"value": {"enabled": enable}}

        _base_api(
            request=f"/v1.0/myorg/groups/{workspace_id}/datasets/{dataset_id}/refreshSchedule",
            method="patch",
            payload=payload,
        )

        print(
            f"{icons.green_dot} Scheduled refresh for the '{dataset_name}' within the '{workspace_name}' workspace has been enabled."
        )


def delete_semantic_model(dataset: str | UUID, workspace: Optional[str | UUID] = None):
    """
    Deletes a semantic model.

    This is a wrapper function for the following API: `Items - Delete Semantic Model <https://learn.microsoft.com/rest/api/fabric/semanticmodel/items/delete-semantic-model>`_.

    Parameters
    ----------
    dataset: str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    delete_item(item=dataset, type="SemanticModel", workspace=workspace)



import sempy.fabric as fabric
import pandas as pd
from sempy_labs._helper_functions import (
    format_dax_object_name,
    resolve_dataset_name_and_id,
    resolve_workspace_name_and_id,
)
from typing import Optional, Tuple
from sempy._utils._log import log
from uuid import UUID


@log
def show_unsupported_direct_lake_objects(
    dataset: str | UUID, workspace: Optional[str | UUID] = None
) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """
    Returns a list of a semantic model's objects which are not supported by Direct Lake based on
    `official documentation <https://learn.microsoft.com/power-bi/enterprise/directlake-overview#known-issues-and-limitations>`_.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame, pandas.DataFrame, pandas.DataFrame
        3 pandas dataframes showing objects in a semantic model which are not supported by Direct Lake.
    """

    pd.options.mode.chained_assignment = None

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    (dataset_name, dataset_id) = resolve_dataset_name_and_id(dataset, workspace_id)

    fabric.refresh_tom_cache(workspace=workspace)

    dfT = fabric.list_tables(dataset=dataset_id, workspace=workspace_id)
    dfC = fabric.list_columns(dataset=dataset_id, workspace=workspace_id)
    dfR = fabric.list_relationships(dataset=dataset_id, workspace=workspace_id)

    # Calc tables
    dfT_filt = dfT[dfT["Type"] == "Calculated Table"]
    dfT_filt.rename(columns={"Name": "Table Name"}, inplace=True)
    t = dfT_filt[["Table Name", "Type"]]

    # Calc columns
    dfC_filt = dfC[(dfC["Type"] == "Calculated") | (dfC["Data Type"] == "Binary")]
    c = dfC_filt[["Table Name", "Column Name", "Type", "Data Type", "Source"]]

    # Relationships
    dfC["Column Object"] = format_dax_object_name(dfC["Table Name"], dfC["Column Name"])
    dfR["From Object"] = format_dax_object_name(dfR["From Table"], dfR["From Column"])
    dfR["To Object"] = format_dax_object_name(dfR["To Table"], dfR["To Column"])
    merged_from = pd.merge(
        dfR, dfC, left_on="From Object", right_on="Column Object", how="left"
    )
    merged_to = pd.merge(
        dfR, dfC, left_on="To Object", right_on="Column Object", how="left"
    )

    dfR["From Column Data Type"] = merged_from["Data Type"]
    dfR["To Column Data Type"] = merged_to["Data Type"]

    dfR_filt = dfR[(dfR["From Column Data Type"] != dfR["To Column Data Type"])]
    r = dfR_filt[
        [
            "From Table",
            "From Column",
            "To Table",
            "To Column",
            "From Column Data Type",
            "To Column Data Type",
        ]
    ]

    # print('Calculated Tables are not supported...')
    # display(t)
    # print("Learn more about Direct Lake limitations here: https://learn.microsoft.com/power-bi/enterprise/directlake-overview#known-issues-and-limitations")
    # print('Calculated columns are not supported. Columns of binary data type are not supported.')
    # display(c)
    # print('Columns used for relationship cannot be of data type datetime and they also must be of the same data type.')
    # display(r)

    return t, c, r



import sempy
import pandas as pd
from typing import List, Optional, Union
from sempy._utils._log import log
import sempy_labs._icons as icons
from sempy_labs._helper_functions import (
    get_language_codes,
    _create_spark_session,
)
from uuid import UUID


@log
def translate_semantic_model(
    dataset: str | UUID,
    languages: Union[str, List[str]],
    exclude_characters: Optional[str] = None,
    workspace: Optional[str | UUID] = None,
) -> pd.DataFrame:
    """
    Translates names, descriptions, display folders for all objects in a semantic model.

    Parameters
    ----------
    dataset : str | uuid.UUID
        Name or ID of the semantic model.
    languages : str, List[str]
        The language code(s) in which to translate the semantic model.
    exclude_characters : str
        A string specifying characters which will be replaced by a space in the translation text when sent to the translation service.
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        Shows a pandas dataframe which displays all of the translations in the semantic model.
    """

    from synapse.ml.services import Translate
    from pyspark.sql.functions import col, flatten
    from sempy_labs.tom import connect_semantic_model

    icons.sll_tags.append("TranslateSemanticModel")

    def _clean_text(text, exclude_chars):
        if exclude_chars:
            for char in exclude_chars:
                text = text.replace(char, " ")
        return text

    if isinstance(languages, str):
        languages = [languages]

    languages = get_language_codes(languages)

    df_prep = pd.DataFrame(
        columns=["Object Type", "Name", "Description", "Display Folder"]
    )

    final_df = pd.DataFrame(columns=["Value", "Translation"])

    with connect_semantic_model(
        dataset=dataset, readonly=False, workspace=workspace
    ) as tom:

        for o in tom.model.Tables:
            oName = _clean_text(o.Name, exclude_characters)
            oDescription = _clean_text(o.Description, exclude_characters)
            new_data = {
                "Name": o.Name,
                "TName": oName,
                "Object Type": "Table",
                "Description": o.Description,
                "TDescription": oDescription,
                "Display Folder": None,
                "TDisplay Folder": None,
            }
            df_prep = pd.concat(
                [df_prep, pd.DataFrame(new_data, index=[0])], ignore_index=True
            )
        for o in tom.all_columns():
            oName = _clean_text(o.Name, exclude_characters)
            oDescription = _clean_text(o.Description, exclude_characters)
            oDisplayFolder = _clean_text(o.DisplayFolder, exclude_characters)
            new_data = {
                "Name": o.Name,
                "TName": oName,
                "Object Type": "Column",
                "Description": o.Description,
                "TDescription": oDescription,
                "Display Folder": o.DisplayFolder,
                "TDisplay Folder": oDisplayFolder,
            }
            df_prep = pd.concat(
                [df_prep, pd.DataFrame(new_data, index=[0])], ignore_index=True
            )
        for o in tom.all_measures():
            oName = _clean_text(o.Name, exclude_characters)
            oDescription = _clean_text(o.Description, exclude_characters)
            oDisplayFolder = _clean_text(o.DisplayFolder, exclude_characters)
            new_data = {
                "Name": o.Name,
                "TName": oName,
                "Object Type": "Measure",
                "Description": o.Description,
                "TDescription": oDescription,
                "Display Folder": o.DisplayFolder,
                "TDisplay Folder": oDisplayFolder,
            }
            df_prep = pd.concat(
                [df_prep, pd.DataFrame(new_data, index=[0])], ignore_index=True
            )
        for o in tom.all_hierarchies():
            oName = _clean_text(o.Name, exclude_characters)
            oDescription = _clean_text(o.Description, exclude_characters)
            oDisplayFolder = _clean_text(o.DisplayFolder, exclude_characters)
            new_data = {
                "Name": o.Name,
                "TName": oName,
                "Object Type": "Hierarchy",
                "Description": o.Description,
                "TDescription": oDescription,
                "Display Folder": o.DisplayFolder,
                "TDisplay Folder": oDisplayFolder,
            }
            df_prep = pd.concat(
                [df_prep, pd.DataFrame(new_data, index=[0])], ignore_index=True
            )
        for o in tom.all_levels():
            oName = _clean_text(o.Name, exclude_characters)
            oDescription = _clean_text(o.Description, exclude_characters)
            new_data = {
                "Name": o.Name,
                "TName": oName,
                "Object Type": "Level",
                "Description": o.Description,
                "TDescription": oDescription,
                "Display Folder": None,
                "TDisplay Folder": None,
            }
            df_prep = pd.concat(
                [df_prep, pd.DataFrame(new_data, index=[0])], ignore_index=True
            )

        spark = _create_spark_session()
        df = spark.createDataFrame(df_prep)

        columns = ["Name", "Description", "Display Folder"]

        for clm in columns:
            columnToTranslate = f"T{clm}"
            translate = (
                Translate()
                .setTextCol(columnToTranslate)
                .setToLanguage(languages)
                .setOutputCol("translation")
                .setConcurrency(5)
            )

            transDF = (
                translate.transform(df)
                .withColumn("translation", flatten(col("translation.translations")))
                .withColumn("translation", col("translation.text"))
                .select("Object Type", clm, columnToTranslate, "translation")
            )

            df_panda = transDF.toPandas()
            df_panda = df_panda[~df_panda[clm].isin([None, ""])][[clm, "translation"]]

            df_panda = df_panda.rename(columns={clm: "value"})
            final_df = pd.concat([final_df, df_panda], ignore_index=True)

        def set_translation_if_exists(object, language, property, index):

            if property == "Name":
                trans = object.Name
            elif property == "Description":
                trans = object.Description
            elif property == "Display Folder":
                trans = object.DisplayFolder

            df_filt = final_df[final_df["value"] == trans]
            if not df_filt.empty:
                translation_value = df_filt["translation"].str[index].iloc[0]
                tom.set_translation(
                    object=object,
                    language=language,
                    property=property,
                    value=translation_value,
                )

        for language in languages:
            index = languages.index(language)
            tom.add_translation(language=language)
            print(
                f"{icons.in_progress} Translating {clm.lower()}s into the '{language}' language..."
            )

            for t in tom.model.Tables:
                set_translation_if_exists(
                    object=t, language=language, property="Name", index=index
                )
                set_translation_if_exists(
                    object=t, language=language, property="Description", index=index
                )
            for c in tom.all_columns():
                set_translation_if_exists(
                    object=c, language=language, property="Name", index=index
                )
                set_translation_if_exists(
                    object=c, language=language, property="Description", index=index
                )
                set_translation_if_exists(
                    object=c, language=language, property="Display Folder", index=index
                )
            for c in tom.all_measures():
                set_translation_if_exists(
                    object=c, language=language, property="Name", index=index
                )
                set_translation_if_exists(
                    object=c, language=language, property="Description", index=index
                )
                set_translation_if_exists(
                    object=c, language=language, property="Display Folder", index=index
                )
            for c in tom.all_hierarchies():
                set_translation_if_exists(
                    object=c, language=language, property="Name", index=index
                )
                set_translation_if_exists(
                    object=c, language=language, property="Description", index=index
                )
                set_translation_if_exists(
                    object=c, language=language, property="Display Folder", index=index
                )
            for c in tom.all_levels():
                set_translation_if_exists(
                    object=c, language=language, property="Name", index=index
                )
                set_translation_if_exists(
                    object=c, language=language, property="Description", index=index
                )

    result = pd.DataFrame(
        columns=[
            "Language",
            "Object Type",
            "Table Name",
            "Object Name",
            "Translated Object Name",
            "Description",
            "Translated Description",
            "Display Folder",
            "Translated Display Folder",
        ]
    )
    with connect_semantic_model(
        dataset=dataset, readonly=True, workspace=workspace
    ) as tom:

        sempy.fabric._client._utils._init_analysis_services()
        import Microsoft.AnalysisServices.Tabular as TOM

        for c in tom.model.Cultures:
            for tr in c.ObjectTranslations:
                oType = str(tr.Object.ObjectType)
                oName = tr.Object.Name
                tValue = tr.Value
                prop = str(tr.Property)

                if tr.Object.ObjectType == TOM.ObjectType.Table:
                    desc = tom.model.Tables[oName].Description
                    new_data = {
                        "Language": c.Name,
                        "Table Name": oName,
                        "Object Name": oName,
                        "Object Type": oType,
                        "Description": desc,
                    }
                    result = pd.concat(
                        [result, pd.DataFrame(new_data, index=[0])], ignore_index=True
                    )
                    condition = (
                        (result["Language"] == c.Name)
                        & (result["Table Name"] == oName)
                        & (result["Object Name"] == oName)
                        & (result["Object Type"] == oType)
                    )
                elif tr.Object.ObjectType == TOM.ObjectType.Level:
                    hierarchyName = tr.Object.Parent.Name
                    tName = tr.Object.Parent.Parent.Name
                    levelName = "'" + hierarchyName + "'[" + oName + "]"
                    desc = (
                        tom.model.Tables[tName]
                        .Hierarchies[hierarchyName]
                        .Levels[oName]
                        .Description
                    )
                    new_data = {
                        "Language": c.Name,
                        "Table Name": tName,
                        "Object Name": levelName,
                        "Object Type": oType,
                        "Description": desc,
                    }
                    result = pd.concat(
                        [result, pd.DataFrame(new_data, index=[0])], ignore_index=True
                    )
                    condition = (
                        (result["Language"] == c.Name)
                        & (result["Table Name"] == tName)
                        & (result["Object Name"] == levelName)
                        & (result["Object Type"] == oType)
                    )
                elif tr.Object.ObjectType == TOM.ObjectType.Column:
                    tName = tr.Object.Table.Name
                    desc = tom.model.Tables[tName].Columns[oName].Description
                    display_folder = (
                        tom.model.Tables[tName].Columns[oName].DisplayFolder
                    )
                    new_data = {
                        "Language": c.Name,
                        "Table Name": tName,
                        "Object Name": oName,
                        "Object Type": oType,
                        "Description": desc,
                        "Display Folder": display_folder,
                    }
                    result = pd.concat(
                        [result, pd.DataFrame(new_data, index=[0])], ignore_index=True
                    )
                    condition = (
                        (result["Language"] == c.Name)
                        & (result["Table Name"] == tName)
                        & (result["Object Name"] == oName)
                        & (result["Object Type"] == oType)
                    )
                elif tr.Object.ObjectType == TOM.ObjectType.Measure:
                    tName = tr.Object.Table.Name
                    desc = tom.model.Tables[tName].Measures[oName].Description
                    display_folder = (
                        tom.model.Tables[tName].Measures[oName].DisplayFolder
                    )
                    new_data = {
                        "Language": c.Name,
                        "Table Name": tName,
                        "Object Name": oName,
                        "Object Type": oType,
                        "Description": desc,
                        "Display Folder": display_folder,
                    }
                    result = pd.concat(
                        [result, pd.DataFrame(new_data, index=[0])], ignore_index=True
                    )
                    condition = (
                        (result["Language"] == c.Name)
                        & (result["Table Name"] == tName)
                        & (result["Object Name"] == oName)
                        & (result["Object Type"] == oType)
                    )
                elif tr.Object.ObjectType == TOM.ObjectType.Hierarchy:
                    tName = tr.Object.Table.Name
                    desc = tom.model.Tables[tName].Hierarchies[oName].Description
                    display_folder = (
                        tom.model.Tables[tName].Hierarchies[oName].DisplayFolder
                    )
                    new_data = {
                        "Language": c.Name,
                        "Table Name": tName,
                        "Object Name": oName,
                        "Object Type": oType,
                        "Description": desc,
                        "Display Folder": display_folder,
                    }
                    result = pd.concat(
                        [result, pd.DataFrame(new_data, index=[0])], ignore_index=True
                    )
                    condition = (
                        (result["Language"] == c.Name)
                        & (result["Table Name"] == tName)
                        & (result["Object Name"] == oName)
                        & (result["Object Type"] == oType)
                    )

                if prop == "Caption":
                    result.loc[condition, "Translated Object Name"] = tValue
                elif prop == "Description":
                    result.loc[condition, "Translated Description"] = tValue
                else:
                    result.loc[condition, "Translated Display Folder"] = tValue
        result.fillna("", inplace=True)

    return result



import pandas as pd
from uuid import UUID
import sempy_labs._icons as icons
from typing import List
from sempy_labs._helper_functions import (
    _is_valid_uuid,
    _base_api,
    _create_dataframe,
)
from sempy._utils._log import log


def resolve_user_id(user: str | UUID) -> UUID:
    """
    Resolves the user ID from the user principal name or ID.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    user : str | uuid.UUID
        The user ID or user principal name.

    Returns
    -------
    uuid.UUID
        The user ID.
    """

    if _is_valid_uuid(user):
        return user
    else:
        result = _base_api(request=f"users/{user}", client="graph").json()
        return result.get("id")


@log
def get_user(user: str | UUID) -> pd.DataFrame:
    """
    Shows properties of a given user.

    This is a wrapper function for the following API: `Get a user <https://learn.microsoft.com/graph/api/user-get>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    user : str | uuid.UUID
        The user ID or user principal name.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing properties of a given user.
    """

    result = _base_api(request=f"users/{user}", client="graph").json()

    new_data = {
        "User Id": result.get("id"),
        "User Principal Name": result.get("userPrincipalName"),
        "User Name": result.get("displayName"),
        "Mail": result.get("mail"),
        "Job Title": result.get("jobTitle"),
        "Office Location": result.get("officeLocation"),
        "Mobile Phone": result.get("mobilePhone"),
        "Business Phones": str(result.get("businessPhones")),
        "Preferred Language": result.get("preferredLanguage"),
        "Surname": result.get("surname"),
    }

    return pd.DataFrame([new_data])


@log
def list_users() -> pd.DataFrame:
    """
    Shows a list of users and their properties.

    This is a wrapper function for the following API: `List users <https://learn.microsoft.com/graph/api/user-list>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing a list of users and their properties.
    """

    result = _base_api(request="users", client="graph").json()

    columns = {
        "User Id": "string",
        "User Principal Name": "string",
        "User Name": "string",
        "Mail": "string",
        "Job Title": "string",
        "Office Location": "string",
        "Mobile Phone": "string",
        "Business Phones": "string",
        "Preferred Language": "string",
        "Surname": "string",
    }

    df = _create_dataframe(columns=columns)

    for v in result.get("value"):
        new_data = {
            "User Id": v.get("id"),
            "User Principal Name": v.get("userPrincipalName"),
            "User Name": v.get("displayName"),
            "Mail": v.get("mail"),
            "Job Title": v.get("jobTitle"),
            "Office Location": v.get("officeLocation"),
            "Mobile Phone": v.get("mobilePhone"),
            "Business Phones": str(v.get("businessPhones")),
            "Preferred Language": v.get("preferredLanguage"),
            "Surname": v.get("surname"),
        }

        df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


@log
def send_mail(
    user: UUID | str,
    subject: str,
    to_recipients: str | List[str],
    content: str,
    content_type: str = "Text",
    cc_recipients: str | List[str] = None,
):
    """
    Sends an email to the specified recipients.

    This is a wrapper function for the following API: `user: sendMail <https://learn.microsoft.com/graph/api/user-sendmail>`_.

    Service Principal Authentication is required (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    user : uuid.UUID | str
        The user ID or user principal name.
    subject : str
        The email subject.
    to_recipients : str | List[str]
        The email address of the recipients.
    content : str
        The email content.
    content_type : str, default="Text"
        The email content type. Options: "Text" or "HTML".
    cc_recipients : str | List[str], default=None
        The email address of the CC recipients.
    """

    if content_type.lower() == "html":
        content_type = "HTML"
    else:
        content_type = "Text"

    user_id = resolve_user_id(user=user)

    if isinstance(to_recipients, str):
        to_recipients = [to_recipients]

    if isinstance(cc_recipients, str):
        cc_recipients = [cc_recipients]

    to_email_addresses = [
        {"emailAddress": {"address": email}} for email in to_recipients
    ]

    cc_email_addresses = (
        [{"emailAddress": {"address": email}} for email in cc_recipients]
        if cc_recipients
        else None
    )

    payload = {
        "message": {
            "subject": subject,
            "body": {
                "contentType": content_type,
                "content": content,
            },
            "toRecipients": to_email_addresses,
        },
    }

    if cc_email_addresses:
        payload["message"]["ccRecipients"] = cc_email_addresses

    _base_api(
        request=f"users/{user_id}/sendMail",
        client="graph",
        status_codes=202,
        payload=payload,
        method="post",
    )

    print(f"{icons.green_dot} The email has been sent to {to_recipients}.")



import pandas as pd


def report_bpa_rules() -> pd.DataFrame:
    """
    Shows the default rules for the report BPA used by the run_report_bpa function.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe containing the default rules for the run_report_bpa function.
    """

    rules = pd.DataFrame(
        [
            (
                "Error Prevention",
                "Semantic Model",
                "Error",
                "Fix report objects which reference invalid semantic model objects",
                lambda df: df["Valid Semantic Model Object"] == False,
                "This rule highlights visuals, report filters, page filters or visual filters which reference an invalid semantic model object (i.e Measure/Column/Hierarchy).",
            ),
            (
                "Performance",
                "Custom Visual",
                "Warning",
                "Remove custom visuals which are not used in the report",
                lambda df: df["Used in Report"] == False,
                "Removing unused custom visuals from a report may lead to faster report performance.",
            ),
            (
                "Performance",
                "Page",
                "Warning",
                "Reduce the number of visible visuals on the page",
                lambda df: df["Visible Visual Count"] > 15,
                'Reducing the number of visable visuals on a page will lead to faster report performance. This rule flags pages with over " + visVisuals + " visible visuals.',
            ),
            (
                "Performance",
                "Visual",
                "Warning",
                "Reduce the number of objects within visuals",
                lambda df: df["Visual Object Count"] > 5,
                "Reducing the number of objects (i.e. measures, columns) which are used in a visual will lead to faster report performance.",
            ),
            (
                "Performance",
                ["Report Filter", "Page Filter", "Visual Filter"],
                "Warning",
                "Reduce usage of filters on measures",
                lambda df: df["Object Type"] == "Measure",
                "Measure filters may cause performance degradation, especially against a large semantic model.",
            ),
            (
                "Performance",
                "Visual",
                "Warning",
                "Avoid setting 'Show items with no data' on columns",
                lambda df: df["Show Items With No Data"],
                "This setting will show all column values for all columns in the visual which may lead to performance degradation.",
                "https://learn.microsoft.com/power-bi/create-reports/desktop-show-items-no-data",
            ),
            (
                "Performance",
                "Page",
                "Warning",
                "Avoid tall report pages with vertical scrolling",
                lambda df: df["Height"] > 720,
                "Report pages are designed to be in a single view and not scroll. Pages with scrolling is an indicator that the page has too many elements.",
            ),
            (
                "Performance",
                "Custom Visual",
                "Info",
                "Reduce usage of custom visuals",
                lambda df: df["Custom Visual Name"] == df["Custom Visual Name"],
                "Using custom visuals may lead to performance degradation.",
            ),
            (
                "Maintenance",
                "Report Level Measure",
                "Info",
                "Move report-level measures into the semantic model.",
                lambda df: df["Measure Name"] == df["Measure Name"],
                "It is a best practice to keep measures defined in the semantic model and not in the report.",
            ),
            (
                "Performance",
                ["Report Filter", "Page Filter", "Visual Filter"],
                "Info",
                "Reduce usage of TopN filtering within visuals",
                lambda df: df["Type"] == "TopN",
                "TopN filtering may cause performance degradation, especially against a high cardinality column.",
            ),
            # ('Performance', 'Custom Visual', 'Warning', "Set 'Edit Interactions' for non-data visuals to 'none'",
            # lambda df: df['Custom Visual Name'] == df['Custom Visual Name'],
            # "Setting 'Edit Interactions' to 'None' for non-data visuals may improve performance (since these visuals do not necessitate interactions between other visuals). 'Edit Interactions' may be found in the 'Format' tab of the ribbon in Power BI Desktop.",
            # )
        ],
        columns=[
            "Category",
            "Scope",
            "Severity",
            "Rule Name",
            "Expression",
            "Description",
            "URL",
        ],
    )

    return rules



import pandas as pd
import sempy_labs._icons as icons
from typing import Optional
from sempy_labs._helper_functions import (
    resolve_workspace_name_and_id,
    resolve_capacity_id,
    _base_api,
    _create_dataframe,
)
from uuid import UUID


def delete_user_from_workspace(
    email_address: str, workspace: Optional[str | UUID] = None
):
    """
    Removes a user from a workspace.

    This is a wrapper function for the following API: `Groups - Delete User In Group <https://learn.microsoft.com/rest/api/power-bi/groups/delete-user-in-group>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    email_address : str
        The email address of the user.
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    _base_api(
        request=f"/v1.0/myorg/groups/{workspace_id}/users/{email_address}",
        method="delete",
        client="fabric_sp",
    )
    print(
        f"{icons.green_dot} The '{email_address}' user has been removed from accessing the '{workspace_name}' workspace."
    )


def update_workspace_user(
    email_address: str,
    role_name: str,
    principal_type: Optional[str] = "User",
    workspace: Optional[str | UUID] = None,
):
    """
    Updates a user's role within a workspace.

    This is a wrapper function for the following API: `Groups - Update Group User <https://learn.microsoft.com/rest/api/power-bi/groups/update-group-user>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    email_address : str
        The email address of the user.
    role_name : str
        The `role <https://learn.microsoft.com/rest/api/power-bi/groups/add-group-user#groupuseraccessright>`_ of the user within the workspace.
    principal_type : str, default='User'
        The `principal type <https://learn.microsoft.com/rest/api/power-bi/groups/add-group-user#principaltype>`_.
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    role_names = icons.workspace_roles
    role_name = role_name.capitalize()
    if role_name not in role_names:
        raise ValueError(
            f"{icons.red_dot} Invalid role. The 'role_name' parameter must be one of the following: {role_names}."
        )
    principal_types = icons.principal_types
    principal_type = principal_type.capitalize()
    if principal_type not in principal_types:
        raise ValueError(
            f"{icons.red_dot} Invalid princpal type. Valid options: {principal_types}."
        )

    payload = {
        "emailAddress": email_address,
        "groupUserAccessRight": role_name,
        "principalType": principal_type,
        "identifier": email_address,
    }

    _base_api(
        request=f"/v1.0/myorg/groups/{workspace_id}/users",
        method="put",
        payload=payload,
        client="fabric_sp",
    )
    print(
        f"{icons.green_dot} The '{email_address}' user has been updated to a '{role_name}' within the '{workspace_name}' workspace."
    )


def list_workspace_users(workspace: Optional[str | UUID] = None) -> pd.DataFrame:
    """
    A list of all the users of a workspace and their roles.

    This is a wrapper function for the following API: `Workspaces - List Workspace Role Assignments <https://learn.microsoft.com/rest/api/fabric/core/workspaces/list-workspace-role-assignments>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe the users of a workspace and their properties.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    columns = {
        "User Name": "string",
        "Email Address": "string",
        "Role": "string",
        "Type": "string",
        "User ID": "string",
    }
    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request=f"v1/workspaces/{workspace_id}/roleAssignments",
        uses_pagination=True,
        client="fabric_sp",
    )

    for r in responses:
        for v in r.get("value", []):
            p = v.get("principal", {})
            new_data = {
                "User Name": p.get("displayName"),
                "User ID": p.get("id"),
                "Type": p.get("type"),
                "Role": v.get("role"),
                "Email Address": p.get("userDetails", {}).get("userPrincipalName"),
            }
            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


def add_user_to_workspace(
    email_address: str,
    role_name: str,
    principal_type: Optional[str] = "User",
    workspace: Optional[str | UUID] = None,
):
    """
    Adds a user to a workspace.

    This is a wrapper function for the following API: `Groups - Add Group User <https://learn.microsoft.com/rest/api/power-bi/groups/add-group-user>`_.

    Parameters
    ----------
    email_address : str
        The email address of the user. Also accepts the user identifier.
    role_name : str
        The `role <https://learn.microsoft.com/rest/api/power-bi/groups/add-group-user#groupuseraccessright>`_ of the user within the workspace.
    principal_type : str, default='User'
        The `principal type <https://learn.microsoft.com/rest/api/power-bi/groups/add-group-user#principaltype>`_.
    workspace : str | uuid.UUID, default=None
        The name or ID of the workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    role_names = icons.workspace_roles
    role_name = role_name.capitalize()
    if role_name not in role_names:
        raise ValueError(
            f"{icons.red_dot} Invalid role. The 'role_name' parameter must be one of the following: {role_names}."
        )
    plural = "n" if role_name == "Admin" else ""
    principal_types = icons.principal_types
    principal_type = principal_type.capitalize()
    if principal_type not in principal_types:
        raise ValueError(
            f"{icons.red_dot} Invalid princpal type. Valid options: {principal_types}."
        )

    payload = {
        "emailAddress": email_address,
        "groupUserAccessRight": role_name,
        "principalType": principal_type,
        "identifier": email_address,
    }

    _base_api(
        request=f"/v1.0/myorg/groups/{workspace_id}/users",
        method="post",
        payload=payload,
    )
    print(
        f"{icons.green_dot} The '{email_address}' user has been added as a{plural} '{role_name}' within the '{workspace_name}' workspace."
    )


def assign_workspace_to_capacity(
    capacity: str | UUID,
    workspace: Optional[str | UUID] = None,
    **kwargs,
):
    """
    Assigns a workspace to a capacity.

    This is a wrapper function for the following API: `Workspaces - Assign To Capacity <https://learn.microsoft.com/rest/api/fabric/core/workspaces/assign-to-capacity>`_.

    Parameters
    ----------
    capacity : str | uuid.UUID
        The name or ID of the capacity.
    workspace : str | uuid.UUID, default=None
        The name or ID of the Fabric workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    if "capacity_name" in kwargs:
        capacity = kwargs["capacity_name"]
        print(
            f"{icons.warning} The 'capacity_name' parameter is deprecated. Please use 'capacity' instead."
        )

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)
    capacity_id = resolve_capacity_id(capacity=capacity)

    payload = {"capacityId": capacity_id}

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/assignToCapacity",
        method="post",
        payload=payload,
        status_codes=[200, 202],
    )
    print(
        f"{icons.green_dot} The '{workspace_name}' workspace has been assigned to the '{capacity}' capacity."
    )


def unassign_workspace_from_capacity(workspace: Optional[str | UUID] = None):
    """
    Unassigns a workspace from its assigned capacity.

    This is a wrapper function for the following API: `Workspaces - Unassign From Capacity <https://learn.microsoft.com/rest/api/fabric/core/workspaces/unassign-from-capacity>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The name or ID of the Fabric workspace.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    _base_api(
        request=f"/v1/workspaces/{workspace_id}/unassignFromCapacity",
        method="post",
        status_codes=[200, 202],
        client="fabric_sp",
    )
    print(
        f"{icons.green_dot} The '{workspace_name}' workspace has been unassigned from its capacity."
    )


def list_workspace_role_assignments(
    workspace: Optional[str | UUID] = None,
) -> pd.DataFrame:
    """
    Shows the members of a given workspace.

    This is a wrapper function for the following API: `Workspaces - List Workspace Role Assignments <https://learn.microsoft.com/rest/api/fabric/core/workspaces/list-workspace-role-assignments>`_.

    Service Principal Authentication is supported (see `here <https://github.com/microsoft/semantic-link-labs/blob/main/notebooks/Service%20Principal.ipynb>`_ for examples).

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.

    Returns
    -------
    pandas.DataFrame
        A pandas dataframe showing the members of a given workspace and their roles.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    columns = {
        "User Name": "string",
        "User Email": "string",
        "Role Name": "string",
        "Type": "string",
    }
    df = _create_dataframe(columns=columns)

    responses = _base_api(
        request=f"v1/workspaces/{workspace_id}/roleAssignments",
        uses_pagination=True,
        client="fabric_sp",
    )

    for r in responses:
        for i in r.get("value", []):
            principal = i.get("principal", {})
            new_data = {
                "User Name": principal.get("displayName"),
                "Role Name": i.get("role"),
                "Type": principal.get("type"),
                "User Email": principal.get("userDetails", {}).get("userPrincipalName"),
            }
            df = pd.concat([df, pd.DataFrame(new_data, index=[0])], ignore_index=True)

    return df


def delete_workspace(workspace: Optional[str | UUID] = None):
    """
    Deletes a workspace.

    This is a wrapper function for the following API: `Workspaces - Delete Workspace <https://learn.microsoft.com/rest/api/fabric/core/workspaces/delete-workspace>`_.

    Parameters
    ----------
    workspace : str | uuid.UUID, default=None
        The Fabric workspace name or ID.
        Defaults to None which resolves to the workspace of the attached lakehouse
        or if no lakehouse attached, resolves to the workspace of the notebook.
    """

    (workspace_name, workspace_id) = resolve_workspace_name_and_id(workspace)

    _base_api(
        request=f"v1/workspaces/{workspace_id}", method="delete", client="fabric_sp"
    )

    print(f"{icons.green_dot} The '{workspace_name}' workspace has been deleted.")
